import threading
import sys
import traceback
import os
import signal

"""
usage: install()

Once installed, all exceptions caught from within threads will cause the program to end.
"""


def sendKillSignal(etype, value, tb):
    lines = traceback.format_exception(etype, value, tb)
    for line in lines:
        print(line, flush=True)

    os.kill(os.getpid(), signal.SIGTERM)


original_init = threading.Thread.__init__


def patched_init(self, *args, **kwargs):
    original_init(self, *args, **kwargs)
    original_run = self.run

    def patched_run(*args, **kw):
        try:
            original_run(*args, **kw)
        except Exception:
            sys.excepthook(*sys.exc_info())
    self.run = patched_run


def install():
    sys.excepthook = sendKillSignal
    threading.Thread.__init__ = patched_init



from dozent.main import Dozent

from dozent.catch_thread_exceptions import install as _catch_thread_exceptions
from dozent.preprocess import Preprocess

_catch_thread_exceptions()

__all__ = ['Dozent', 'Preprocess']



import argparse
import datetime
import json
import multiprocessing
import os
import time
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import List, Dict, Any

from dozent.downloader_tools import DownloaderTools
from dozent.progress_tracker import ProgressTracker

CURRENT_FILE_PATH = Path(__file__)
DEFAULT_DATA_DIRECTORY = CURRENT_FILE_PATH.parent.parent / 'data'
TWITTER_ARCHIVE_STREAM_LINKS_PATH = CURRENT_FILE_PATH.parent / 'twitter-archive-stream-links.json'

FIRST_DAY_OF_SUPPORT = datetime.date(2017, 6, 1)
LAST_DAY_OF_SUPPORT = datetime.date(2020, 6, 30)

class _DownloadWorker(Thread):  # skip_tests

    def __init__(self, queue: Queue, download_dir: Path, tracker: ProgressTracker = None):
        Thread.__init__(self)
        self.queue = queue
        self.download_dir = download_dir
        self.tracker = tracker

    def run(self):
        while True:
            # Get the work from the queue and expand the tuple
            link = self.queue.get()
            try:
                DownloaderTools.download_with_pysmartdl(link, str(self.download_dir), tracker=self.tracker)
            finally:
                self.queue.task_done()


class Dozent:
    # TODO: Move start_date and end_date as optional arguments
    __instance__ = None

    def __init__(self):
        if Dozent.__instance__ is None:
            Dozent.__instance__ = self
            with open(TWITTER_ARCHIVE_STREAM_LINKS_PATH) as file:
                self.date_links: List[Dict[str, str]] = json.loads(file.read())

        else:
            raise RuntimeError(f"Singleton {self.__class__.__name__} class is created more than once!")

    @staticmethod
    def _make_date_from_date_link(date_link: Dict[str, str]) -> datetime.date:
        """
        Function to make a date given the date link. For months that don't have individual dates (date = 'NaN'), we
        assume that it's the first.
        :param date_link: date dictionary from `data/test_sample_files.json`. This is used to convert to the date
        :return: date object for the associated month
        """
        day = int(date_link['day']) if date_link['day'] != 'NaN' else 1
        month = int(date_link['month'])
        year = int(date_link['year'])

        return datetime.date(year=year, month=month, day=day)

    def get_links_for_days(self, start_date: datetime.date, end_date: datetime.date) -> List:
        """
        Function to get the links for the given start and end days
        :return: date dictionaries that are within self.start_date and self.end_dates
        """
        links = []

        if not FIRST_DAY_OF_SUPPORT >= start_date:
            RuntimeError(f'We currently only support the range {FIRST_DAY_OF_SUPPORT.strftime("%d, %b %Y")}, '
                         f'{LAST_DAY_OF_SUPPORT.strftime("%d, %b %Y")}'
                         f'\nWe\'re planning on adding support for it soon.'
                         f'Need that data sooner? Add an issue to out GitHub repo and we\'ll '
                         f'walk you through the process'
                         f'Link to our GitHub '
                         f'Issues: https://github.com/Twitter-Public-Analysis/Twitter-Public-Analysis/issues')

        elif not end_date <= end_date:
            RuntimeError(f'We currently only support the range {FIRST_DAY_OF_SUPPORT.strftime("%d, %b %Y")}, '
                         f'{LAST_DAY_OF_SUPPORT.strftime("%d, %b %Y")}'
                         f'\nWe\'re planning on adding support for it soon.'
                         f'Need that data sooner? Add an issue to out GitHub repo and we\'ll '
                         f'walk you through the process'
                         f'Link to our GitHub '
                         f'Issues: https://github.com/Twitter-Public-Analysis/Twitter-Public-Analysis/issues')

        for date_link in self.date_links:
            date = Dozent._make_date_from_date_link(date_link)
            if start_date <= date <= end_date:
                links.append(date_link)

        return links

    def download_timeframe(self, start_date: datetime.date, end_date: datetime.date, verbose: bool = True,
                           download_dir: Path = DEFAULT_DATA_DIRECTORY):
        """
        Download all tweet archives from self.start_date to self.end_date
        :return: None
        """

        # Create a queue to communicate with the worker threads
        queue = Queue()
        if verbose:
            tracker = ProgressTracker()
            tracker.daemon = True
            tracker.start()
        else:
            tracker = None

        os.makedirs(download_dir, exist_ok=True)

        for x in range(multiprocessing.cpu_count()):
            worker = _DownloadWorker(queue, download_dir, tracker=tracker)
            # worker.set_verbosity(verbose=verbosity)
            # Setting daemon to True will let the main thread exit even though the workers are blocking
            worker.daemon = True
            worker.start()

        for sample_date in self.get_links_for_days(start_date=start_date, end_date=end_date):
            print(f"Queueing tweets download for {sample_date['month']}-{sample_date['year']}")
            queue.put(sample_date['link'])

        queue.join()
        tracker.join()


def main(command_line_arguments: Dict[str, Any]):  # skip_tests
    _start_time = time.time()
    verbose = not command_line_arguments['quiet']
    _dozent_object = Dozent()

    _dozent_object.download_timeframe(start_date=command_line_arguments['start_date'],
                                      end_date=command_line_arguments['end_date'],
                                      verbose=verbose)

    if command_line_arguments['timeit']:
        print(f"Download Time: {datetime.timedelta(seconds=(time.time() - _start_time))}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='A powerful downloader to get tweets from twitter for our compute. '
                                                 'The first step of many')
    parser.add_argument('-s', '--start-date', help="The date from where we download. The format must be: YYYY-MM-DD",
                        required=True, type=lambda s: datetime.datetime.strptime(s, '%Y-%m-%d').date())
    parser.add_argument('-e', '--end-date', help="The last day that we download. The format must be: YYYY-MM-DD",
                        required=True, type=lambda s: datetime.datetime.strptime(s, '%Y-%m-%d').date())
    parser.add_argument('-t', '--timeit', help='Show total program runtime', default=True)
    parser.add_argument('-o', '--output-directory', help='Output Directory where the file will be stored. '
                                                         'Defaults to the data/ directory',
                        default=DEFAULT_DATA_DIRECTORY)
    parser.add_argument('-q', '--quiet', help='Turn off output (except for errors and warnings)', action='store_true')
    args = parser.parse_args()
    main(vars(args))



import os
import shutil
import tarfile
import zipfile

from smpamorpheus.data_loading import DataLoading
from pathlib import Path
from typing import Union, List


class Preprocess:
    __instance__ = None
    _output_formats = ['csv', 'json', 'parquet', 'sql']
    _supported_formats = ['csv', 'json']
    _unsupported_formats = ['parquet', 'sql']

    def __init__(self):
        if Preprocess.__instance__ is None:
            Preprocess.__instance__ = self
        else:
            raise RuntimeError(f"Singleton {self.__class__.__name__} class is created more than once!")

    @staticmethod
    def __untar_file(file_path: Union[str, Path], destination: Union[str, Path], verbose: bool): # skip_tests
        """
        Untars a single tar file into target directory
        :return: None
        """
        if verbose:
            print(f"Extracting {file_path}")
        tar_file = tarfile.open(file_path)
        try:
            tar_file.extractall(destination)
        except:
            tar_file.close()
        tar_file.close()

    @staticmethod
    def __unzip_file(file_path: Union[str, Path], destination: Union[str, Path], verbose: bool): # skip_tests
        """
        Unzips a single zip file into target directory
        :return: None
        """
        if verbose:
            print(f"Extracting {file_path}")
        with zipfile.ZipFile(file_path, 'r') as zip_ref:
            try:
                zip_ref.extractall(destination)
            except:
                zip_ref.close()
            zip_ref.close()

    @staticmethod
    def extract_directory(directory_path: Union[str, Path], destination: Union[str, Path], verbose: bool = True,
                          delete_archive: bool = False):
        """
        Extracts all files in a directory into target directory
        :param delete_archive:
        :param directory_path: path of the directory
        :param destination: where the files will be stored
        :param verbose: show output?
        :return: None
        """

        if not os.path.isdir(destination):
            os.mkdir(destination)

        files = [f for f in os.listdir(directory_path) if os.path.isfile(os.path.join(directory_path, f))]
        for file in files:
            file_extension = file[-3:]
            if file_extension == 'tar':
                if verbose:
                    print(f"\nUntaring: {file}")
                Preprocess.__untar_file(f"{directory_path}/{file}", destination, verbose)
            elif file_extension == 'zip':
                if verbose:
                    print(f"\nUnzipping: {file}")
                Preprocess.__unzip_file(f"{directory_path}/{file}", destination, verbose)
            else:
                raise RuntimeError(f"File extension .{file_extension} not recognized.")
        if verbose:
            print(f"\nSuccessfully extracted all files in {directory_path} to {destination}")

        if delete_archive:
            shutil.rmtree(directory_path)

    @staticmethod
    def store_tweets_to_file_format(directory_path: Union[str, Path], destination: Union[str, Path],
                                    suffix: str = '*.json*',
                                    recursive: bool = True,
                                    output_format='csv') -> List[str]:
        """
        Extracting all tweets and storing them as csv.
        This function recursively searches the `directory_path` for files that match the given `suffix`

        :param directory_path: path of the directory where the data is stored.
        :param suffix: suffix: suffix to search for files with
        :param destination: where the output needs to be stored
        :param recursive: recursively search for files?
        :param output_format: the format that the data should be stored in.
                              While the possible values are:  ['csv', 'json', 'parquet', 'sql'], we currently test for
                              only ['csv', 'json']
        :return: returns the list of all of the output files
        """
        if output_format not in Preprocess._supported_formats:
            raise ValueError(f'The given format: {output_format} is not in {Preprocess._supported_formats}')

        files_list = DataLoading.get_files_list(pathname=directory_path, suffix=suffix, recursive=recursive)
        data = DataLoading.get_twitter_data_from_file_list(file_lst=files_list, remove_deleted_tweets=True)
        data = data.to_dataframe()

        if output_format == 'csv':
            return data.to_csv(destination)

        elif output_format == 'json':
            return data.to_json(destination)

        elif output_format == 'parquet':  # skip_tests
            raise NotImplementedError('Parquet has not been yet supported because of the added dependency of '
                                      'fastparque/pyarrow, which we might look into later'
                                      'If you would like this feature, please make an issue here: '
                                      'https://github.com/Twitter-Public-Analysis/Twitter-Public-Analysis/issues')

        elif output_format == 'sql':  # skip_tests
            raise NotImplementedError('SQL has not been yet implemented. '
                                      'If you would like this feature, please make an issue here: '
                                      'https://github.com/Twitter-Public-Analysis/Twitter-Public-Analysis/issues')

        else:
            raise ValueError(f'The given format: {output_format} is not recognized')


if __name__ == "__main__":
    pass



import multiprocessing
import os
import time
from typing import Tuple

from pySmartDL import SmartDL

try:
    from dozent.progress_tracker import ProgressTracker
except ModuleNotFoundError:
    from .progress_tracker import ProgressTracker


class DownloaderTools:
    __instance__ = None

    def __init__(self):
        if DownloaderTools.__instance__ is None:
            DownloaderTools.__instance__ = self

        else:
            raise RuntimeError(f"Singleton {self.__class__.__name__} class is created more than once!")

    @staticmethod
    def _make_progress_status(downloader_obj: SmartDL) -> Tuple[float, str, str]:
        """Function to make progress bar

        :param downloader_obj: SmartDL object that's currently downloading a file
        :return: the progress percentage in [0,100] and a string prefix/suffix to be output before/after a progress bar.
        """

        status = downloader_obj.get_status()
        filesize = downloader_obj.get_final_filesize() >> 20

        try:
            dl_size = downloader_obj.get_dl_size() >> 20
            speed = downloader_obj.get_speed(human=True)
            eta = downloader_obj.get_eta(human=True)
            progress_percentage = 100.0 * downloader_obj.get_progress()

            eta = eta.split(',')[0]
            eta_parts = eta.split(' ')
            eta_unit = eta_parts[1][:3]
            if eta_unit == 'hou':
                eta_unit = 'hrs'
            eta = f"{float(eta_parts[0]):3.0f}{eta_unit}"

        except AttributeError:
            dl_size = 0
            speed = ''
            eta = ''
            progress_percentage = 0

        prefix = f"[{status}] {dl_size}Mb/{filesize}Mb {f'@{speed}' if speed else ''}"
        suffix = f"[{int(progress_percentage):3.0f}%{f', {eta} left' if eta else ''}]"

        return progress_percentage, prefix, suffix

    @classmethod
    def download_with_pysmartdl(cls, link: str, download_dir: str, tracker: ProgressTracker = None):
        """
        Downloads file from link using PySmartDL

        :param link: link that needs to be downloaded
        :param download_dir: A relative path to the download directory
        :param tracker: Used to track progress, defaults to None
        """

        downloader_obj = SmartDL(link, download_dir, progress_bar=False)

        task_id = None
        if tracker:
            task_id = tracker.register_task(link, lambda: DownloaderTools._make_progress_status(downloader_obj))

        downloader_obj.start(blocking=False)

        while not downloader_obj.isFinished():
            if task_id is not None:
                tracker.update(task_id)
            time.sleep(.25)

    @classmethod
    def download_with_axel(cls, link: str):  # skip_tests: Only possible on Ubuntu and depreciated
        """Downloads file from link using axel

        :param link: link that needs to be downloaded
        :return: None
        """
        os.system(f"axel --verbose --alternate --num-connections={cls._connections_count} {link}")

    @classmethod
    def download_with_aria2(cls, link: str):  # skip_tests: Only possible on Ubuntu and depreciated
        """
        Downloads file from link using aria2
        :param link: link that needs to be downloaded
        """
        os.system(f"aria2c -x {cls._connections_count} {link}")

    _connections_count = 2 * multiprocessing.cpu_count()

    _downloaders = {
        'pySmartDL': download_with_pysmartdl.__func__,
        'Axel': download_with_axel.__func__,
        'aria2': download_with_aria2.__func__
    }



from distutils.core import setup
setup(
  name = 'dozent',
  packages = ['dozent'],
  version = '0.1',
  license='MIT',
  description = 'Dozent is a powerful downloader that is used to download a ton of twitter data from the internet archive.',
  author = 'Ali Abbas, Eric Burt, Keelin Becker-Wheeler',
  author_email = 'eric.burt@protonmail.com',
  url = 'https://github.com/Social-Media-Public-Analysis/dozent',
  download_url = 'https://github.com/Social-Media-Public-Analysis/dozent/archive/v_02.tar.gz',
  keywords = ['TWITTER', 'SCRAPER', 'DOWNLOAD'],
  install_requires=[
            'pySmartDL',
            'aria2p',
            'pytest',
            'pyfiglet',
            'smpamorpheus',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.6',
  ],
)


import sys
import os

import numpy as np
import threading
import shutil
import queue

from typing import Dict, Callable, Tuple
from collections import namedtuple

Task = namedtuple('Task', 'id progress_callback progress prefix suffix line_length')


class ProgressTracker(threading.Thread):
    """It is assumed that no other printing to stdout will occur while running this ProgressTracker thread.
    """

    def __init__(self):
        threading.Thread.__init__(self)

        self.tasks: Dict[int, Task] = {}
        self.message_queue = queue.Queue()
        self.sum_progress = 0
        self.next_task = 0

        self.terminal_size = None

        resized = self.refresh_terminal_size()
        assert(resized)  # This won't work if we couldn't read the terminal size

        self.lock = threading.Lock()
        self.tasks_pending = False
        self.running = True

    def refresh_terminal_size(self) -> bool:
        """
        Records any change in the terminal size.
        :returns: True if a change is detected, else False
        """

        old_size = self.terminal_size
        try:
            self.terminal_size = os.get_terminal_size()[0]
        except OSError:
            self.terminal_size = shutil.get_terminal_size()[0]

        return old_size != self.terminal_size

    def join(self):
        self.message_queue.join()
        self.stop()

    def stop(self):
        self.running = False
        self.update(0)

    def update(self, task_id: int):
        self.message_queue.put(task_id)

    def register_task(self, name: str, progress_callback: Callable[[], Tuple[float, str, str]]) -> int:
        """Register a new task with the given progress callback.

        :param name: A unique name for this task.
        :param progress_callback: A function, which will report a tuple of:
                                    a progress percentage in [0,100],
                                    a string prefix to be output before the progress bar,
                                    and a string suffix to be output after the progress bar.
        :return: The ID of the registered task.
        """

        task_id = self.next_task
        self.next_task += 1

        # Perform operation in helper function on own thread, so that this function does not block
        thread = threading.Thread(target=self._register_task_helper, args=(name, task_id, progress_callback))
        thread.daemon = True
        thread.start()

        return task_id

    def _register_task_helper(self, name: str, task_id: int, progress_callback: Callable[[], Tuple[float, str]]):
        with self.lock:
            # Create the task, with initial progress 0
            self.tasks[task_id] = (Task(id=task_id,
                                        progress_callback=progress_callback,
                                        progress=0,
                                        prefix=None,
                                        suffix=None,
                                        line_length=0))
            self.tasks_pending = True
            self._delete_progress_bars()
            print(f"Queueing tweet download {task_id}: {name}", flush=True)

    def _delete_progress_bars(self, starting_id: int = 0):
        if self.tasks:
            sys.stdout.write('\b' * (np.sum([self.tasks[i].line_length for i in range(starting_id, len(self.tasks))])))

    def _update_task_progress(self, task_id: int, draw: bool = True):
        """Update the progress bar for the given task

        :param task_id: Identifier of the task to update.
        :param draw: Whether to draw the updated progress bar, defaults to True
        """

        # Update the progress value
        task = self.tasks[task_id]
        assert(task.id == task_id)

        progress, prefix, suffix = task.progress_callback()
        assert(0 <= progress <= 100)

        self.sum_progress += progress - task.progress
        task = Task(id=task_id,
                    progress_callback=task.progress_callback,
                    progress=progress,
                    prefix=prefix,
                    suffix=suffix,
                    line_length=task.line_length)
        self.tasks[task_id] = task

        if draw:
            # Redraw all bars if the terminal size changes
            if self.refresh_terminal_size():
                task_id = 0

            # Return to start of the progress bar for task_id
            self._delete_progress_bars(starting_id=task_id)
            self._draw_task_progress(task_id)

    def _draw_task_progress(self, task_id: int):
        """Updates the progress bar for task_id and after.

        Assumes the console begins at the end of any previously drawn progress bars.
        """

        # We re-write the progress for all tasks, in order to finish at the end of all progress bars (our assumption)
        for task in [self.tasks[i] for i in range(task_id, len(self.tasks))]:
            if task.prefix is None or task.suffix is None:
                continue

            prefix = f"{task.id:3.0f}: "

            bar = ' '*len(prefix)
            bar_width = self.terminal_size - len(prefix) - 2
            assert(bar_width > 0)

            num_dashes = int(task.progress * bar_width / 100.0)
            bar += f"[{'-' * num_dashes}{' ' * (bar_width - num_dashes)}]"

            prefix += task.prefix[:int(self.terminal_size/2)]
            suffix = task.suffix[:int(self.terminal_size/3)]
            spaces = ' ' * (self.terminal_size - len(suffix) - len(prefix))

            line = f"{prefix}{spaces}{suffix}{bar}"

            assert(self.tasks[task.id] == task)
            self.tasks[task.id] = Task(id=task.id,
                                       progress_callback=task.progress_callback,
                                       progress=task.progress,
                                       prefix=task.prefix,
                                       suffix=task.suffix,
                                       line_length=len(line))

            sys.stdout.write(line)

        sys.stdout.flush()

    def _run_task(self):
        task_id = self.message_queue.get(True)

        try:
            with self.lock:
                if self.tasks_pending:
                    self._update_task_progress(0, draw=False)
                    self.tasks_pending = False

                else:
                    self._update_task_progress(task_id, draw=True)

        finally:
            self.message_queue.task_done()

    def run(self):
        while self.running:
            self._run_task()

        self._delete_progress_bars()
        print('')
